#include <LCAPIIOAcc.h>
#include <LC3GlobalsDMA.h>
#include <LC3GV.h>

typedef uint32_t LcIOAccCopyBitOffs;
typedef uint8_t LcIOAccCopyBitSize;

typedef struct
{
  uint32_t  bitOffset;
  uint8_t  bitSize;
  void* variable;
} LcIOAccCopyEntry;

typedef struct
{
  uint32_t   copyEntryCnt;
  LcIOAccCopyEntry const* copyEntries;
} LcIOAccCopyTable;

typedef struct
{
  char const* serviceName;
  LcIOAccServiceId serviceId;
} LcIOAccService;

static LcIOAccCopyEntry ioCopyEntries_1_INPUT[] = 
{
  {0, 1, &(g_LC3_GVInstanceData.LC_VD_I_1)},
};

static LcIOAccCopyTable ioCopyTable_1_INPUT = {1, ioCopyEntries_1_INPUT};

static LcIOAccCopyEntry ioCopyEntries_1_OUTPUT[] = 
{
  {560, 1, &(g_LC3_GVInstanceData.LC_VD_O_1)},
  {573, 1, &(g_LC3_GVInstanceData.LC_VD_O_14)},
  {571, 1, &(g_LC3_GVInstanceData.LC_VD_O_12)},
  {569, 1, &(g_LC3_GVInstanceData.LC_VD_O_10)},
  {570, 1, &(g_LC3_GVInstanceData.LC_VD_O_11)},
  {572, 1, &(g_LC3_GVInstanceData.LC_VD_O_13)},
  {561, 1, &(g_LC3_GVInstanceData.LC_VD_O_2)},
  {562, 1, &(g_LC3_GVInstanceData.LC_VD_O_3)},
  {563, 1, &(g_LC3_GVInstanceData.LC_VD_O_4)},
  {564, 1, &(g_LC3_GVInstanceData.LC_VD_O_5)},
  {565, 1, &(g_LC3_GVInstanceData.LC_VD_O_6)},
  {566, 1, &(g_LC3_GVInstanceData.LC_VD_O_7)},
  {567, 1, &(g_LC3_GVInstanceData.LC_VD_O_8)},
  {568, 1, &(g_LC3_GVInstanceData.LC_VD_O_9)},
};

static LcIOAccCopyTable ioCopyTable_1_OUTPUT = {14, ioCopyEntries_1_OUTPUT};

static LcIOAccCopyTable ioCopyTable_1_MEMORY = {0, NULL};

#define BITS_READ(src,offs,cnt)       ((src) >> (offs) & 0xff >> (8-cnt))
#define BITS_WRITE(dst,offs,cnt,value)    \
  do{                                     \
    (dst) &= ~(BITS_READ(0xff,0,cnt) << (offs)); \
    (dst) |= (value) << (offs); \
  } while (0)

#define ADD_BIT_OFFS(ptr,offs,delta) \
  do{ \
    (offs) += (delta); \
    (ptr) += (offs) / 8; \
    (offs) %= 8; \
  } while(0)

static void copyBits(uint8_t* dst, uint8_t dstBitOffs, uint8_t const* src, uint8_t srcBitOffs, uint8_t bitCnt)
{
  uint8_t bitCntCopy;
  uint8_t bitCntSrc;
  uint8_t bitCntDst;
  uint8_t copyValue;

  while (bitCnt)
  {
    bitCntCopy = bitCnt;
    bitCntSrc = 8 - srcBitOffs;
    if(bitCntSrc < bitCntCopy)
    {
      bitCntCopy = bitCntSrc;
    }
    bitCntDst = 8 - dstBitOffs;
    if(bitCntDst < bitCntCopy)
    {
      bitCntCopy = bitCntDst;
    }
    copyValue = BITS_READ(*src, srcBitOffs, bitCntCopy);
    BITS_WRITE(*dst, dstBitOffs, bitCntCopy, copyValue);
    ADD_BIT_OFFS(dst, dstBitOffs, bitCntCopy);
    ADD_BIT_OFFS(src, srcBitOffs, bitCntCopy);
    bitCnt -= bitCntCopy;
  }
}

static void copyIOsToVariables(LcIOAccCopyTable const* copyTable, uint8_t const* ioSegment)
{
  LcIOAccCopyEntry const* copyEntry;
  LcIOAccCopyEntry const* copyEntryEnd = copyTable->copyEntries + copyTable->copyEntryCnt;

  for (copyEntry = copyTable->copyEntries; copyEntry < copyEntryEnd; copyEntry++)
  {
    copyBits(copyEntry->variable, 0, ioSegment+copyEntry->bitOffset/8, copyEntry->bitOffset % 8,copyEntry->bitSize);
  }
}

static void copyIOsFromVariables(LcIOAccCopyTable const* copyTable, uint8_t* ioSegment)
{
  LcIOAccCopyEntry const* copyEntry;
  LcIOAccCopyEntry const* copyEntryEnd = copyTable->copyEntries + copyTable->copyEntryCnt;

  for (copyEntry = copyTable->copyEntries; copyEntry < copyEntryEnd; copyEntry++)
  {
    copyBits(ioSegment + copyEntry->bitOffset / 8, copyEntry->bitOffset % 8, copyEntry->variable, 0, copyEntry->bitSize);
  }
}

LcIOAccCopyRc lc_ioacc_copyInputData(
  LcIOAccServiceId serviceId, uint8_t const* ioInputData, LcIOAccSegmentSize inputBitSize, LcIOAccAdditionalData *additionalData)
{
  switch(serviceId)
  {
  case 1:
    if(inputBitSize < 1)
    {
      return LC_IOACC_COPY_RC_SEGMENT_SIZE_ERR;
    }
    copyIOsToVariables(&ioCopyTable_1_INPUT, ioInputData);
    return LC_IOACC_COPY_RC_OK;
  }
  return LC_IOACC_COPY_RC_INV_SERVICE_ID;
}

LcIOAccCopyRc lc_ioacc_copyOutputData(
  LcIOAccServiceId serviceId, uint8_t* ioOutputData, LcIOAccSegmentSize outputBitSize, LcIOAccAdditionalData* additionalData)
{
  switch(serviceId)
  {
  case 1:
    if(outputBitSize < 574)
    {
      return LC_IOACC_COPY_RC_SEGMENT_SIZE_ERR;
    }
    copyIOsFromVariables(&ioCopyTable_1_OUTPUT, ioOutputData);
    return LC_IOACC_COPY_RC_OK;
  }
  return LC_IOACC_COPY_RC_INV_SERVICE_ID;
}

LcIOAccCopyRc lc_ioacc_copyMemoryDataToVariables(
  LcIOAccServiceId serviceId, uint8_t const* ioInputData, LcIOAccSegmentSize inputBitSize, LcIOAccAdditionalData *additionalData)
{
  switch(serviceId)
  {
  case 1:
    if(inputBitSize < 0)
    {
      return LC_IOACC_COPY_RC_SEGMENT_SIZE_ERR;
    }
    copyIOsToVariables(&ioCopyTable_1_MEMORY, ioInputData);
    return LC_IOACC_COPY_RC_OK;
  }
  return LC_IOACC_COPY_RC_INV_SERVICE_ID;
}

LcIOAccCopyRc lc_ioacc_copyMemoryDataFromVariables(
  LcIOAccServiceId serviceId, uint8_t* ioOutputData, LcIOAccSegmentSize outputBitSize, LcIOAccAdditionalData* additionalData)
{
  switch(serviceId)
  {
  case 1:
    if(outputBitSize < 0)
    {
      return LC_IOACC_COPY_RC_SEGMENT_SIZE_ERR;
    }
    copyIOsFromVariables(&ioCopyTable_1_MEMORY, ioOutputData);
    return LC_IOACC_COPY_RC_OK;
  }
  return LC_IOACC_COPY_RC_INV_SERVICE_ID;
}

